%{
/**************************************************************/
/* Filename: scanner.l                                        */
/* Purpose:  a Lex specification for a small set of tokens    */
/* Usage: flex scanner.l ; gcc lex.yy.c -lfl ;		      */
/* 		cat test.cf | ./a.out			      */
/* Phase 1 submission for Nick Toothman			      */
/* CS 450, 10/13/2009					      */
/**************************************************************/

#define TRUE 1 
#define FALSE 0 
#define MAXTOKENLEN 40 

#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>

typedef enum 
    /* book-keeping tokens */
   { ERROR, CERROR, FSTRINGERROR, COMMENTERROR, CHARERROR,
     PCHARERROR, ERRORNUM,
    /* reserved words */
    MAIN,BREAK,CLASS,CONST,ELSE,IF,NEW,READ,RETURN,VOID,WHILE,WRITE,
    /* datatypes, constants, comments */
    INT,ID,DTYPE,NUM,FP,FORTRANSTRING,CSTRING,CHAR,PCHAR,COMMENT,
    /* operators */
    PLUS,MINUS,MULT,DIV,MOD,EQ,INEQ,LT,LTE,GT,GTE,AND,OR,ASSIGN,
    INCR,DECR,SEMI,COMMA,PERIOD,LPAR,RPAR,LBRK,RBRK,LBRC,RBRC
   } TokenType;

/* data structure created for symbol table */
struct symbol
{
  char lexeme[255];
  char value[255];
  char type[255];
};

/* fixed size array to hold symbols */
struct symbol symbols[255];
int line_cnt=1;
int col_cnt=1;
int token_cnt=0;
int symbol_cnt=0;


%}

letter		[a-zA-Z]
digit       	[0-9]
number      	{digit}+
floatingpoint	[+-]?{number}?(\.{number})?([eE][+-]?{number})?
identifier	[a-zA-Z_]+[a-zA-Z0-9_]*
newline     	\n
cr		\r
whitespace  	[ \t]+
fortranstring	{number}+H[a-zA-Z0-9 \!\.\?]+
cstring		\"[a-zA-Z0-9 \t\n\\\"\.\!\?\']+\"
comment		\*([^*]|[\r\n]|(\*+([^*/]|[\r\n])))*\*+
char		[a-zA-Z0-9\!\"\#\$\%\&\'\(\)\*\+\,\-\.\/\:\;\<\=\>\?\@\[\\\]\^\_\`\{\|\}\~]
sq		\'
printableChar	{sq}{char}{sq}

%%

"main"		{return MAIN;}
"break"		{return BREAK;}
"class"		{return CLASS;}
"const"		{return CONST;}
"else"          {return ELSE;}
"if"            {return IF;}
"new"		{return NEW;}
"read"          {return READ;}
"return"	{return RETURN;}
"void"		{return VOID;}
"while"		{return WHILE;}
"write"         {return WRITE;}
"int"		{return DTYPE;}
"char"		{return DTYPE;}
"+"		{return PLUS;}
"-"		{return MINUS;}
"*"		{return MULT;}
"/"		{return DIV;}
"%"		{return MOD;}
"=="		{return EQ;}
"!="		{return INEQ;}
"<"		{return LT;}
"<="		{return LTE;}
">"		{return GT;}
">="		{return GTE;}
"&&"		{return AND;}
"||"		{return OR;}
"="		{return ASSIGN;}
"++"		{return INCR;}
"--"		{return DECR;}
";"		{return SEMI;}
","		{return COMMA;}
"."		{return PERIOD;}
"("		{return LPAR;}
")"		{return RPAR;}
"["		{return LBRK;}
"]"		{return RBRK;}
"{"		{return LBRC;}
"}"		{return RBRC;}
{comment}	{return COMMENT;}
{number}        {return NUM;}
{identifier}    {return ID;}
{floatingpoint} {return FP;}
{fortranstring} {return FORTRANSTRING;}
{cstring}	{return CSTRING;}
{printableChar}	{return PCHAR;}
{char}		{return CHAR;}
{newline}       {line_cnt++; col_cnt=1;}
{whitespace}    {col_cnt++; /* skip whitespace */}

%%

/* returns string describing the token type */
char * tokenString(TokenType token)
{
  if (token == ERROR)	return "ERROR";
  
  if (token == FSTRINGERROR)
			return "Invalid Fortran string";
  if (token == COMMENTERROR)
			return "Illegally nested comments";
  if (token == ERRORNUM)return "Overflow of integer constant";
  if (token == PCHARERROR)
			return "Illegal character constant";
  if (token == CHARERROR)
			return "Illegal character";
  if (token == CLASS ||
      token == DTYPE)	return "Keyword";
  if (token <= WRITE)	return "Keyword";
  if (token == INT)	return "Integer";
  if (token == ID)	return "Identifier";
  if (token == NUM)	return "Number";
  if (token == FP)      return "Floating Point"; 
  if (token == FORTRANSTRING)
			return "Fortran String";
  if (token == CSTRING)	return "C-string";
  if (token == PCHAR)	return "Printable Character";
  if (token == COMMENT)	return "Comment";
  if (token <= RBRC)	return "Operator";
}

/* returns the appropriate "real" token given the value
and the assumed token */
TokenType ValidateToken(TokenType token, char *value)
{
  /* check that numbers are below INT_MAX by testing cast */
  if (token == NUM)
  {
    int test = atoi(value);
    if (test == INT_MAX || test == INT_MIN)
      return ERRORNUM;
    else return NUM;
  }

  /* check that comments are not illegally nested */
  if (token == COMMENT)
  {
    
    int i;
    
    for(i = 2; i < strlen(value)-1; i++)
    {
      if (value[i] == '/')
      {
	if (value[i+1] == '*')
	  return COMMENTERROR;
      }
    }

    if (value[i] == '/') return COMMENT;
  }

  /* checks that Fortran-style strings are formatted properly
     note: allows for strings of length > 10 and allows for
     spaces in the string literal! */
  if (token == FORTRANSTRING)
  {
    char c, *orig;
    int num = 0, count = 0, hFound = 0;

    orig = yytext;

    do
    {
      c = *orig;

      if (hFound) count++;
      
      else if (c == 'H' && !hFound) hFound = 1;

      orig++;
    } while (c != '\0');

    count--;
    num = atoi(yytext);

    if (num == count) return FORTRANSTRING;
    else return FSTRINGERROR;
  }

  /* checks that printable characters are properly delimited
     on both sides by single quotes */
  if (token == PCHAR)
  {
    if (value[0] == '\'' && value[strlen(yytext)-1] == '\'')
      return PCHAR;
    else return PCHARERROR;
  }

  /* if a character is found by itself, see that it's not
     acting as an identifier (make sure it's legal) */
  if (token == CHAR)
  {
    char c = *yytext;
    if (!isalnum(c) || c != '_')
      return CHARERROR;
  }

  return token;
}

/* traverses symbol table and makes sure that the symbol
to be added has not yet been added */
int symbolExists(const char * lexeme, const char * value)
{
  int index = 0;

  for(index = 0; index < symbol_cnt; index++)
  {
    if (strncmp(symbols[index].lexeme, lexeme, strlen(lexeme)) == 0
      && strncmp(symbols[index].value, value, strlen(value)) == 0)
    {
      return index;
    }
  }

  return -1;
}

/* handles determination of token type based on syntax and
content */
TokenType getToken(void)
{ static int firstTime = TRUE;
  TokenType currentToken;

  // initialization 
  if (firstTime)
  {   firstTime = FALSE;
      line_cnt++;
      yyin = stdin;
      yyout = stdout;
      printf("\n");
  }

  currentToken = (TokenType) yylex();

  if (!(currentToken)) return currentToken;

  /* all tokens get validated */
  currentToken = ValidateToken(currentToken, yytext);

  /* these tokens have the potential to be added to
     the symbol table */
  if(
      currentToken == ID ||
      currentToken == NUM ||
      currentToken == FP ||
      currentToken == FORTRANSTRING ||
      currentToken == CSTRING ||
      currentToken == PCHAR
    )
  {
    if (symbolExists(tokenString(currentToken), yytext) == -1)
    {
      strncpy(symbols[symbol_cnt].lexeme, tokenString(currentToken), 255);
      strncpy(symbols[symbol_cnt].value, yytext, 255);
      symbol_cnt++;
    }
  }

  col_cnt += yyleng;
  token_cnt++;

  /* explicitly define what tokens get printed
     and reported on */
  if (currentToken == CLASS ||
      currentToken == ID ||
      currentToken == INT ||
      currentToken == SEMI ||
      currentToken == CHAR ||
      currentToken == LBRC ||
      currentToken == LPAR ||
      currentToken == RPAR ||
      currentToken == ASSIGN ||
      currentToken == RBRC ||
      currentToken == NUM ||
      currentToken == DTYPE
      )
  {
    printf("line %d, col %d: ", line_cnt, col_cnt);
    printf("%s \t %s\n", tokenString(currentToken), yytext);
  }

  /* handle errors for various types */
  if (currentToken == FSTRINGERROR)
  {
    printf("line %d, col %d: ", line_cnt, col_cnt);
    printf("ERROR: Hollerith notation not met\n");
  }

  if (currentToken == COMMENTERROR)
  {
    printf("line %d, col %d: ", line_cnt, col_cnt);
    printf("ERROR: illegally nested comments\n");
  }

  if (currentToken == PCHARERROR)
  {
    printf("line %d, col %d: ", line_cnt, col_cnt);
    printf("ERROR: missing ' on some side of char constant %s\n", yytext);
  }

  if (currentToken == CHARERROR)
  {
    printf("line %d, col %d: ", line_cnt, col_cnt);
    printf("ERROR: %c is not a valid character\n", yytext[0]);
  }

  return currentToken;
}

int main()
{
  /* parse input, get tokens */
  while( getToken() != 0 );
  printf("\nTOTAL TOKEN CNT: %d\n",token_cnt);

  /* traverse symbol tree, print results */
  int whatever = 0;
  printf("Symbol Table\nIndex %*s Value %*s Lexeme\n", 
    15, "", 15, "");
  for(whatever = 0; whatever < symbol_cnt; whatever++)
  {
    printf("%d %*s %*s\n", 
      whatever, 25,
      symbols[whatever].value, 25,
      symbols[whatever].lexeme );
  }

  return 0;

}
